/*

    Copyright (C) 2013  Hong Jen Yee (PCMan) <pcman.tw@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/


#include "view.h"
#include <libfm-qt/filemenu.h>
#include <libfm-qt/foldermenu.h>
#include "application.h"
#include "settings.h"
#include "application.h"
#include "mainwindow.h"
#include "launcher.h"
#include <QAction>
#include <libfm-qt/core/terminal.h>
#include <libfm-qt/fileoperation.h>
#include <QStandardPaths>
#include <QGraphicsDropShadowEffect>
#include <QApplication>
#include <QClipboard>
#include <QProcess>
#include <QDBusInterface>

namespace NDEFileMAN {

View::View(Fm::FolderView::ViewMode _mode, QWidget* parent):
    Fm::FolderView(_mode, parent),
    isDesktopPrepare_(false),
    isFolderReload_(false) {

    Settings& settings = static_cast<Application*>(qApp)->settings();
    updateFromSettings(settings);
}

View::~View() {
}

void View::onFileClicked(int type, const std::shared_ptr<const Fm::FileInfo>& fileInfo) {
    if(type == MiddleClick) {
        if(fileInfo->isDir()) {
            Q_EMIT openDirRequested(fileInfo->path(), OpenInNewTab);
        }
    }
    else {
        Fm::FolderView::onFileClicked(type, fileInfo);
    }
}

void View::onNewWindow() {
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    // FIXME: open the files in a new window
    Application* app = static_cast<Application*>(qApp);
    app->openFolders(menu->files());
}

void View::onNewTab() {
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    for(auto& file: files) {
        Fm::FilePath path = file->path();
        QString computerPath = file->path().uri().get();
        if((computerPath.mid(0,12) == QString("computer:///")) && file->isMountable()) {
            path = Fm::FilePath::fromLocalPath(file->target().c_str());
        }
        Q_EMIT openDirRequested(path, OpenInNewTab);
    }
}

void View::onOpenInTerminal() {
    Application* app = static_cast<Application*>(qApp);
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    for(auto& file: files) {
        app->openFolderInTerminal(file->path());
    }
}

void View::onSearch() {

}

void View::prepareFileMenu(Fm::FileMenu* menu) {
    Application* app = static_cast<Application*>(qApp);
    menu->setConfirmDelete(app->settings().confirmDelete());
    menu->setConfirmTrash(app->settings().confirmTrash());
    //menu->setUseTrash(app->settings().useTrash());

    // add some more menu items for dirs
    bool all_native = true;
    bool all_directory = true;
    bool isMultipleFile = false;    // 是否多选
    bool isDefault = false;         // 是否是桌面的默认应用
    auto files = menu->files();
    for(auto& fi: files) {
        if(!fi->isDir()) {
            all_directory = false;
        }
        else if(fi->isDir() && !fi->isNative()) {
            all_native = false;
        }
        QString name = QString::fromStdString(fi->name());
        if((name == QLatin1String("trash-can.desktop"))
            || (name == QLatin1String("user-home.desktop"))
            || (name == QLatin1String("computer.desktop"))
            || (name == QLatin1String("network.desktop"))) {
            isDefault = true;
         }
    }
    // 如果选中包含有默认图标，则删除部分右键菜单功能
    if(isDefault) {
        menu->removeAction(menu->openWithMenuAction());
        menu->removeAction(menu->cutAction());
        menu->removeAction(menu->copyAction());
        menu->removeAction(menu->pasteAction());
        menu->removeAction(menu->deleteAction());
        menu->removeAction(menu->compressAction());
        menu->removeAction(menu->compressToAction());
        menu->removeAction(menu->renameAction());
    }
    // 如果是多选，删除右键菜单中的重命名功能
    if (files.size() > 1) {
        isMultipleFile = true;
        menu->removeAction(menu->renameAction());
    }

    bool allTrash =  menu->allTrash();      // 是否是在回收站中
    bool allComputer = menu->allComputer();   // 是否时在计算机中

    QAction* action;
    if((!isDesktopPrepare_ && all_directory && !allTrash && !isDefault) || allComputer ) {
        action = new QAction(QIcon::fromTheme("window-new"), tr("Open in New T&ab"), menu);
        connect(action, &QAction::triggered, this, &View::onNewTab);
        menu->insertAction(menu->separator1(), action);

        action = new QAction(QIcon::fromTheme("window-new"), tr("Open in New Win&dow"), menu);
        connect(action, &QAction::triggered, this, &View::onNewWindow);
        menu->insertAction(menu->separator1(), action);

        // TODO: add search
        // action = menu->addAction(_("Search"));

        // if(all_native) {
        //     action = new QAction(QIcon::fromTheme("utilities-terminal"), tr("Open in Termina&l"), menu);
        //     connect(action, &QAction::triggered, this, &View::onOpenInTerminal);
        //     menu->insertAction(menu->propertiesAction(), action);
        // }
    }

    if (!allTrash && !allComputer && !isDefault) {
        // 判断是否是打开到应用是否在白名单或是否为文件夹
        QFile norootfile;
        bool isRoot = true;
        QString filedir = QString("/usr/share/nde-fileman/nde/notsupporroot.list");
        norootfile.setFileName(filedir);
        QTextStream in(&norootfile);
        if (norootfile.open(QIODevice::ReadOnly | QIODevice::Text)) {
            QString t;
            for(auto& file: files) {
                while(!in.atEnd()) {
                    t =in.readLine();
                    if( t == QString::fromStdString(file->name()) || file->isDir()) {
                        isRoot = false;
                    }
                }
            }
        }

        // 以管理员取权限运行应用
        if (isRoot && !isMultipleFile){
            action = new QAction(tr("Open as Root"), menu);
            connect(action, &QAction::triggered, this, &View::onOpenAsRootTriggered);
            menu->insertAction(menu->separator1(), action);
        }

        // 右键图标创建快捷方式
        action = new QAction(tr("Create Shortcut"), menu);
        connect(action, &QAction::triggered, this, &View::onCreateShortcutWithTriggered);
        menu->insertAction(menu->separator1(), action);

        // 右键图标发送到
        action = new QAction(tr("Copy to"), menu);
        // 创建 "Copy to" 子菜单
        action->setMenu(createCopytoMenuSub(menu));
        menu->insertAction(menu->separator1(), action);
        // 判断是否有粘贴,如果无设置为不可用
        if(menu->pasteAction()){
            QClipboard* clipboard = QApplication::clipboard();
            const QMimeData* data = clipboard->mimeData();
            Fm::FilePathList paths;
            bool isCut = false;
            std::tie(paths, isCut) = Fm::parseClipboardData(*data);
            if(paths.empty()) {
                menu->pasteAction()->setEnabled(false);
                menu->pasteAction()->setIcon(QIcon(":/res/nde-paste-disable.png"));
            }
        }
    }
}

void View::prepareFolderMenu(Fm::FolderMenu* menu) {
    // 刷新
    QAction* action = menu->addAction(tr("Reload"));
    action->setIcon(QIcon(":/res/nde-reload.png"));
    connect(action, &QAction::triggered, this, &View::onFolderReload);
    menu->insertAction(menu->pasteAction(), action);
    
    // 隐藏回收站右键菜单中的“创建”功能和添加“清空回收站”功能
    auto path = menu->view()->path();
    if (path == Fm::FilePath::fromUri("trash:///")) {
        menu->removeAction(menu->createAction());
        menu->removeAction(menu->createFolderAction());
        menu->removeAction(menu->pasteAction());
        
        Settings& settings = static_cast<Application* >(qApp)->settings();
        action = menu->addAction(tr("Empty Trash"));
        // disable the item is Trash is empty
        auto items = menu->view()->model()->rowCount();
        if (items == 0) {
            action->setEnabled(false);
        }
        // empty Trash on clicking the item
        connect(action, &QAction::triggered, this, [this, &settings] {
            Fm::FilePathList files;
            files.push_back(Fm::FilePath::fromUri("trash:///"));
            Fm::FileOperation::deleteFiles(std::move(files), settings.confirmDelete(), this);
        });
        menu->insertAction(menu->propertiesAction(), action);
    }
    else if (path == Fm::FilePath::fromUri("computer:///")) {
        menu->removeAction(menu->createAction());
        menu->removeAction(menu->createFolderAction());
        menu->removeAction(menu->pasteAction());
    }
    else {
        // 右键菜单打开终端
        action = new QAction(tr("Terminal"), menu);
        action->setIcon(QIcon(":/res/nde-terminal.png"));
        connect(action, &QAction::triggered, this, &View::onOpenFolderTerminal);
        menu->insertAction(menu->separator2(), action);
    }

    // 判断是否有粘贴,如果无设置为不可用
    if(menu->pasteAction()){
        QClipboard* clipboard = QApplication::clipboard();
        const QMimeData* data = clipboard->mimeData();
        Fm::FilePathList paths;
        bool isCut = false;
        std::tie(paths, isCut) = Fm::parseClipboardData(*data);
        if(paths.empty()) {
            menu->pasteAction()->setEnabled(false);
            menu->pasteAction()->setIcon(QIcon(":/res/nde-paste-disable.png"));
        }
    }
    
    // 隐藏“全选”，“反选”，“显示隐藏文件”功能
    menu->removeAction(menu->selectAllAction());
    menu->removeAction(menu->invertSelectionAction());
    menu->removeAction(menu->showHiddenAction());
    // 文件管理器查看功能信号接收
    connect(menu, &Fm::FolderMenu::DirectoryViewSignals, this, &View::DirectoryViewSlots);
}

void View::updateFromSettings(Settings& settings) {

    setIconSize(Fm::FolderView::IconMode, QSize(settings.bigIconSize(), settings.bigIconSize()));
    setIconSize(Fm::FolderView::CompactMode, QSize(settings.smallIconSize(), settings.smallIconSize()));
    setIconSize(Fm::FolderView::ThumbnailMode, QSize(settings.thumbnailIconSize(), settings.thumbnailIconSize()));
    setIconSize(Fm::FolderView::DetailedListMode, QSize(settings.smallIconSize(), settings.smallIconSize()));

    setMargins(settings.folderViewCellMargins());

    setAutoSelectionDelay(settings.autoSelectionDelay());

    Fm::ProxyFolderModel* proxyModel = model();
    if(proxyModel) {
        proxyModel->setShowThumbnails(settings.showThumbnails());
        proxyModel->setBackupAsHidden(settings.backupAsHidden());
    }
}

// 右键图标创建快捷方式
void View::onCreateShortcutWithTriggered() {
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    for(auto& file: files) {
        QString filepath = file->path().parent().localPath().get();
        QString filename = file->displayName();
        QString newName = tr("%1 - ").arg(filename) + tr("A Shortcut");
        if(file->isDesktopEntry()) {
            Fm::FilePathList paths;
            paths.push_back(file->path());
            if(!paths.empty()) {
                Fm::copyFilesToClipboard(paths);
            }
            Fm::pasteFilesFromClipboard(path());
        }
        else {
            filepath = filepath + tr("/") + newName;
            Fm::FileOperation::symlinkFile(file->path(),Fm::FilePath::fromLocalPath(filepath.toLocal8Bit().data()));
        }
    }
}

// 右键菜单打开终端
void View::onOpenFolderTerminal() {
    Application* app = static_cast<Application*>(qApp);
    Fm::FolderMenu* menu = static_cast<Fm::FolderMenu*>(sender()->parent());
    auto path = menu->view()->path();
    app->openFolderInTerminal(path);
}

// 以管理员权限运行应用
void View::onOpenAsRootTriggered() {
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();

    Application* app = static_cast<Application*>(qApp);
    Settings& settings = app->settings();
    for(auto& file: files) {
        if(!settings.suCommand().isEmpty()) {
            // run the su command
            // FIXME: it's better to get the filename of the current process rather than hard-code nde-fileman here.
            QByteArray suCommand = settings.suCommand().toLocal8Bit();
            QByteArray programCommand = app->applicationFilePath().toLocal8Bit();
            programCommand += " %U";

            // if %s exists in the su command, substitute it with the program
            int substPos = suCommand.indexOf("%s");
            if(substPos != -1) {
                // replace %s with program
                suCommand.replace(substPos, 2, programCommand);
            }
            else {
                /* no %s found so just append to it */
                suCommand += programCommand;
            }

            Fm::GAppInfoPtr appInfo{g_app_info_create_from_commandline(suCommand.constData(), nullptr, GAppInfoCreateFlags(0), nullptr), false};

            if(appInfo) {
                auto cwd = file->path();
                Fm::GErrorPtr err;
                auto uri = cwd.uri();
                GList* uris = g_list_prepend(nullptr, uri.get());

                if(!g_app_info_launch_uris(appInfo.get(), uris, nullptr, &err)) {
                    QMessageBox::critical(this, tr("Error"), QString::fromUtf8(err->message));
                }

                g_list_free(uris);
            }
        }
        else {
            // show an error message and ask the user to set the command
            QMessageBox::critical(this, tr("Error"), tr("Switch user command is not set."));
            app->preferences("advanced");
        }
    }
}

void View::onCopyToHomeTriggered() {
    QString home_path = QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    Fm::copyFilesToClipboard(files.paths());
    Fm::pasteFilesFromClipboard(Fm::FilePath::fromPathStr(home_path.toLatin1().data()));
}

void View::onCopyToDesktopTriggered() {
    QString desktop_path = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    Fm::copyFilesToClipboard(files.paths());
    Fm::pasteFilesFromClipboard(Fm::FilePath::fromPathStr(desktop_path.toLatin1().data()));
}

QMenu* View::createCopytoMenuSub(Fm::FileMenu* menu) {
    QMenu* copytoMenu = new QMenu(menu);
    QAction* copyto = new QAction(tr("Home"), menu);
    copyto->setIcon(QIcon(":/res/nde-home.png"));
    connect(copyto, &QAction::triggered, this, &View::onCopyToHomeTriggered);
    copytoMenu->addAction(copyto);
    copyto = new QAction(tr("Desktop"), menu);
    copyto->setIcon(QIcon(":/res/nde-desktop.png"));
    connect(copyto, &QAction::triggered, this, &View::onCopyToDesktopTriggered);
    copytoMenu->addAction(copyto);
    copyto = new QAction(tr("Desktop A Shortcut"), menu);
    connect(copyto, &QAction::triggered, this, &View::onDesktopShortcutTriggered);
    copytoMenu->addAction(copyto);
    // menu样式
    copytoMenu->setWindowFlag(Qt::FramelessWindowHint);
    copytoMenu->setAttribute(Qt::WA_TranslucentBackground);
    bool open = false;
    QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                     QString::fromLatin1("/NdeSession"),
                                                     QString::fromLatin1("org.nde.session"),
                                                     QDBusConnection::sessionBus());
    if(compositer->isValid())
    {
        QDBusMessage reply = compositer->call("getCompositorOpen");
        if(!reply.arguments().isEmpty()) open = reply.arguments().first().toBool();
    }
    if(open)
    {
        copytoMenu->setStyleSheet(QLatin1String("QMenu {background-color: #ffffff; border-radius: 4px; color: #5d626d; margin: 2px;padding: 2px;}"
						"QMenu::item {padding-left: 17px; padding-right: 35px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                                "QMenu::item:selected{background-color:#2c9dfc; color:#ffffff;}"));
    }
    else {
        copytoMenu->setStyleSheet(QLatin1String("QMenu {background-color: #ffffff;border: 1px solid #d8d8d8;  border-radius: 0px; color: #5d626d; padding: 2px;}"
						"QMenu::item {padding-left: 17px; padding-right: 35px; height: 32px; font-family: SourceHanSansCN-Normal; font-size: 14px; text-align:left; color: #5d626d; }"
                                                "QMenu::item:selected{background-color:#2c9dfc; color:#ffffff;}"));
    }
    // 阴影
    QGraphicsDropShadowEffect  * shadow_effect = new QGraphicsDropShadowEffect(this);
    shadow_effect->setOffset(0,0);
    QColor color("#000000");
    color.setAlphaF(0.2);
    shadow_effect->setColor(color);
    shadow_effect->setBlurRadius(4);
    copytoMenu->setGraphicsEffect(shadow_effect);

    return copytoMenu;
}
// 右键菜单“刷新”功能
void View::onFolderReload() {
    isFolderReload_ = true;
    Fm::FolderMenu* menu = static_cast<Fm::FolderMenu*>(sender()->parent());
    menu->view()->folder()->reload();
}

void View::DirectoryViewSlots(int mode) {
    switch(mode) {
    case IconView: {
        this->setFolderViewMode(Fm::FolderView::IconMode);
        break;
    }
    case CompactView: {
        this->setFolderViewMode(Fm::FolderView::CompactMode);
        break;
    }
    case DetailedList: {
        this->setFolderViewMode(Fm::FolderView::DetailedListMode);
        break;
    }
    case ThumbnailView: {
        this->setFolderViewMode(Fm::FolderView::ThumbnailMode);
        break;
    }
    default:
        ;
    }
}

// 创建快捷方式到桌面
void View::onDesktopShortcutTriggered() {
    Fm::FileMenu* menu = static_cast<Fm::FileMenu*>(sender()->parent());
    auto files = menu->files();
    QString desktopPath = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    for(auto& file: files) {
        QString filepath = file->path().parent().localPath().get();
        QString filename = file->displayName();
        QString newName = tr("%1 - ").arg(filename) + tr("A Shortcut");
        if(file->isDesktopEntry()) {
            Fm::FilePathList paths;
            paths.push_back(file->path());
            if(!paths.empty()) {
                Fm::copyFilesToClipboard(paths);
            }
            Fm::pasteFilesFromClipboard(Fm::FilePath::fromPathStr(desktopPath.toLatin1().data()));
        }
        else {
            desktopPath = desktopPath + tr("/") + newName;
            Fm::FileOperation::symlinkFile(file->path(),Fm::FilePath::fromLocalPath(desktopPath.toLocal8Bit().data()));
        }
    }
}

void View::setFolderViewMode(Fm::FolderView::ViewMode mode) {
    Q_EMIT folderViewMode(mode);
}

} // namespace NDEFileMAN
